#if !defined(__cocc_h)
#define __cocc_h

#include "arrays.h"

#if !defined(FAST_STEM)
class COccCol : public DArray<ulong>
{
public:
     COccCol():DArray<ulong>(5, 20) { }
     COccCol(const COccCol &Other):DArray<ulong>(Other) { }
};
#else
class COccColData : public DArray<ulong>
{
public:
    int RefCount;

    COccColData():DArray<ulong>(5, 20), RefCount(1) { }
    COccColData(const COccColData &Other):DArray<ulong>(Other), RefCount(1) { }
};

class COccCol
{
public:
    COccCol();
    COccCol(const COccCol &Other);
    COccCol(short aMaster);
    COccCol(const COccCol &Other, short aMaster);
    virtual ~COccCol();
    COccCol &operator = (const COccCol &Other);

    int GetCount() const { return Data->GetCount(); }
    // N. B.: does _not_ copy shared data
    void Add(ulong Occ);
    ulong operator[](int Index) const { return Data->operator[](Index); }

    short IsMaster() const { return Master; }

private:
    int operator == (const COccCol &Other) const;
    int operator != (const COccCol &Other) const;
    int operator < (const COccCol &Other) const;

private:
    COccColData *Data;
    short Master;
};
#endif

class CComposedOcc
{
private:
	DString OccWord;
	COccCol OccCol;

	CComposedOcc(CComposedOcc &Src); // not implemented
	CComposedOcc &operator = (CComposedOcc &Src);
		// not implemented

public:
	CComposedOcc(DString &aOccWord);
     CComposedOcc(const DString &aOccWord, const COccCol &aOccCol);

	~CComposedOcc() { }
		// not virtual to save space;
		// do not inherit from CComposedOcc!

	void Add(ulong Occ);

	const DString &GetWord() const { return OccWord; }

	// count is out param;
	// count == 0 => returns NULL
	ulong *AllocOccArray(unsigned short &Count);
	int GetOccCount() { return OccCol.GetCount(); }

#if defined(NOCPCONSTRUCT)
	int operator == (const CComposedOcc *Other)
		{ return ::operator ==(GetWord(), Other->GetWord()); }
	int operator < (const CComposedOcc *Other)
		{ return GetWord() < Other->GetWord(); }
	int operator > (const CComposedOcc *Other)
		{ return GetWord() > Other->GetWord(); }
#else
	int operator == (const CComposedOcc &Other) const;
	int operator < (const CComposedOcc &Other) const;
#endif
};

#endif

