#include "myfndt.h"
#include "wcache.h"
#include "wctestm.h"

#if defined(DEBUG_CWordCache)
#include <stdio.h>
#include <string.h>
#endif

const int CWordCache::DefaultLimit = 1500;
/* profile 28. 12. 96:
   500           22 min
   1500          20 min 
   3000          20 min 

   with YankNeighbor:
   1500          12 min 
   3000          13 min */

const int CWordCache::DefaultMaxMemberOccCount = 200;

CWordCache::CWordCache(int aLimit,
    int aMaxMemberOccCount):
    CacheArr(aLimit, 0),
    FirstKeyAlias(NULL),
    LastKeyAlias(NULL),
    MaxMemberOccCount(aMaxMemberOccCount),
    LastYankedIndex(INT_MAX)
{
}

CComposedOcc *CWordCache::NewMerge(CSimpleOcc *InOcc)
{
    CObjPtr<CSimpleOcc> SimpleOcc = InOcc;
#if defined(DEBUG_CWordCache)
    static int n = 0;
//    printf("%d: %s\n", n, SimpleOcc->GetWord().c_str());
    n++;
    if (FirstKeyAlias)
	   Assert(FindIndex(*FirstKeyAlias) != INT_MAX);
#endif
    CWordCacheTestMember TestMember(SimpleOcc->GetWord());
    int Index = CacheArr.Find(&TestMember);
    if (Index != INT_MAX)
    {    CWordCacheMember *Member = static_cast<CWordCacheMember *>(
	            CacheArr[Index]);
         Member->Merge(SimpleOcc);
	    if (Member->OccCount() >= MaxMemberOccCount)
	    {    Detach(Member->GetKey(), Index);
		    CComposedOcc *OutOcc = Member->DropOcc();
		    delete Member;
		    LastYankedIndex = Index;
		    return OutOcc;
	    }
	    else
	         return NULL;
    }
    else
    {    CObjPtr<CComposedOcc> OldComposedOcc = NULL;
	    if (CacheArr.IsFull())
	    {    Assert(LastKeyAlias != NULL);
		    OldComposedOcc = Yank(*LastKeyAlias);
	    }

	    CString *OldFirst = FirstKeyAlias;
#if defined(DEBUG_CWordCache)
	    CheckArr();
#endif
	    CacheArr.Add(new CWordCacheMember(SimpleOcc, NULL, OldFirst));
#if defined(DEBUG_CWordCache)
	    CheckArr();
#endif
	    int Kludge = CacheArr.Find(&TestMember);
	    Assert(Kludge != INT_MAX);
	    CWordCacheMember *Member = static_cast<CWordCacheMember *>(CacheArr[
                 Kludge]);
         FirstKeyAlias = const_cast<CString *>(&(Member->GetKey()));
#if defined(DEBUG_CWordCache)
	    Assert(FirstKeyAlias != NULL);
	    Assert(FindIndex(*FirstKeyAlias) != INT_MAX);
#endif

	    if (LastKeyAlias)
	    {   int OldFirstIdx = FindIndex(*OldFirst);
		   Assert(OldFirstIdx != INT_MAX);
		   static_cast<CWordCacheMember *>(CacheArr[
                     OldFirstIdx])->PrevKeyAlias = FirstKeyAlias;
	    }
	    else
	        LastKeyAlias = FirstKeyAlias;

         return OldComposedOcc.Drop();
    }
}

CComposedOcc *CWordCache::YankFirst()
{
    if (CacheArr.IsEmpty())
	   return NULL;
    return Yank(CacheArr[0]->GetKey());
}

CComposedOcc *CWordCache::YankNeighbor()
{
    if (LastYankedIndex == INT_MAX ||
	   LastYankedIndex >= CacheArr.GetCount())
	   return NULL;
    return Yank(CacheArr[LastYankedIndex]->GetKey());
}

int CWordCache::IsEmpty()
{
    return FirstKeyAlias == NULL;
}

int CWordCache::FindIndex(const CString &Key)
{
#if defined(DEBUG_CWordCache)
    const char *p = Key.c_str();
#endif
    CWordCacheTestMember TestMember(Key);
    return CacheArr.Find(&TestMember);
}

CWordCacheMember *CWordCache::Detach(const CString &Key, int Idx)
{   CWordCacheMember *Member;

    Assert(FirstKeyAlias != NULL && LastKeyAlias != NULL);
        // they're NULL only when CacheArr empty

    if (Key == *FirstKeyAlias)
	   FirstKeyAlias = static_cast<CWordCacheMember *>(CacheArr[
                Idx])->NextKeyAlias;
    if (Key == *LastKeyAlias)
	   LastKeyAlias = static_cast<CWordCacheMember *>(CacheArr[
                Idx])->PrevKeyAlias;

    if ((Member = GetPrev(Idx)) != NULL)
	    SkipNext(Member);
    if ((Member = GetNext(Idx)) != NULL)
	    SkipPrev(Member);

    Member = static_cast<CWordCacheMember *>(CacheArr[Idx]);
    CacheArr.DetachByIndex(Idx);
    return Member;
}

CComposedOcc *CWordCache::Yank(const CString &Key)
{
	int FindIdx = FindIndex(Key);
	if (FindIdx == INT_MAX)
		return NULL;

	CObjPtr<CWordCacheMember> Member = Detach(Key, FindIdx);
	CComposedOcc *ComposedOcc = Member->DropOcc();
     LastYankedIndex = FindIdx;
	return ComposedOcc;
}

CWordCacheMember *CWordCache::GetPrev(int Idx)
{
	CWordCacheMember *Member = static_cast<CWordCacheMember *>(CacheArr[Idx]);
	if (!Member->PrevKeyAlias)
		return NULL;

	int PrevIdx = FindIndex(*(Member->PrevKeyAlias));
	Assert(PrevIdx != INT_MAX);
	return static_cast<CWordCacheMember *>(CacheArr[PrevIdx]);
}

CWordCacheMember *CWordCache::GetNext(int Idx)
{
	CWordCacheMember *Member = static_cast<CWordCacheMember *>(CacheArr[Idx]);
	if (!Member->NextKeyAlias)
		return NULL;

	int NextIdx = FindIndex(*(Member->NextKeyAlias));
	Assert(NextIdx != INT_MAX);
	return static_cast<CWordCacheMember *>(CacheArr[NextIdx]);
}

void CWordCache::SkipPrev(CWordCacheMember *Member)
{
	if (!Member->PrevKeyAlias)
		return;

	int PrevIdx = FindIndex(*(Member->PrevKeyAlias));
	Assert(PrevIdx != INT_MAX);
	Member->PrevKeyAlias = static_cast<CWordCacheMember *>(CacheArr[
         PrevIdx])->PrevKeyAlias;
}

void CWordCache::SkipNext(CWordCacheMember *Member)
{
	if (!Member->NextKeyAlias)
		return;

	int NextIdx = FindIndex(*(Member->NextKeyAlias));
	Assert(NextIdx != INT_MAX);
	Member->NextKeyAlias = static_cast<CWordCacheMember *>(CacheArr[
         NextIdx])->NextKeyAlias;
}

#if defined(DEBUG_CWordCache)
void CWordCache::CheckArr()
{
    DISArray<CWordCacheAbstractMember>::DIterator Itr = CacheArr.InitIterator();
    CString PrevKey = "";
//    printf("***\n");
    while (CacheArr.IsDereferenceable(Itr))
    {    CWordCacheAbstractMember *Current = CacheArr.GetCurrent(Itr);
	    const char *Key = Current->GetKey().c_str();
//	    printf("%s\n", Key);
	    if (!PrevKey.is_null())
	    {    const char *p = PrevKey.c_str();
		    Assert(strcmp(p, Key) < 0);
	    }
	    PrevKey = Current->GetKey();
	    ++Itr;
    }
}
#endif

