#include "myfndt.h"
#include "formdisp.h"
#include <stdio.h>
#include "mirror.h"
#include "result.h"
#include "grazeopt.h"
#include "docum.h"
#include "messform.h"
#include "txtcolor.h"
#include "relevanc.h"
#include "clot.h"
#include "clotcurs.h"
#include "passover.h"
#include "puparse.h"

#include "odisptxt.h"
#include "ocolors.h"
#include "omirror.h"
#include "ooutform.h"
#include "oreader.h"
#include "opunct.h"

//#define DEBUG_passage

CFormDisplay::CFormDisplay(CMirror &aMirror,
    CResult &aResult,
    CGrazerOptions &aOptions):
    Mirror(aMirror),
    Result(aResult),
    Options(aOptions),
    OutFile(NULL)
{
    DString OutFileName = Options.Get(O_OUTFILE);
    if (!OutFileName.IsEmpty())
	   OutFile.Open(OutFileName, "w+b");

    SkipClots = IsEmpty(O_CLOTHEAD_1) && IsEmpty(O_CLOTBREAK_1) &&
	       IsEmpty(O_CLOTTAIL_1) && IsEmpty(O_HIGHPASS_4) &&
           IsEmpty(O_LOWPASS_4);

    DString Garbage = aOptions.Get(O_SEPARATORS) + aOptions.Get(O_PUNCT);
    PunctParse = new CPunctParse(Garbage);
}

CFormDisplay::~CFormDisplay()
{
    delete PunctParse;
}

void CFormDisplay::ErrPrint(const DString &Buffer)
{
    printf(Buffer.c_str());
}

void CFormDisplay::OutPrint(const DString &Buffer)
{
    if (Buffer.IsEmpty())
        return;

#if !defined(__GNUG__)
    if (OutFile.IsAssigned())
	   OutFile.Print(Buffer.c_str());
    else
        printf(Buffer.c_str());
#else
// printf/sprintf on long (thousands of characters) strings messes up memory
// (gcc 2.7.2.1 with libc.so.5)
    int Size = Buffer.GetLength() + 1;
    CMemPtr<char> Kludge(Size);
    memcpy((char *)Kludge, Buffer.c_str(), Size);
    int StepSize = 200;
    int From = 0;
    while (From + StepSize < Size)
    {   char c = *(((char *)Kludge) + From + 200);
        *(((char *)Kludge) + From + 200) = 0;
        OutPrintStep(((char *)Kludge) + From);
        *(((char *)Kludge) + From + 200) = c;
        From += StepSize;
    }
    OutPrintStep(((char *)Kludge) + From); 
#endif
}

#if defined(__GNUG__)
void CFormDisplay::OutPrintStep(const char *Buffer)
{
    if (OutFile.IsAssigned())
	OutFile.Print(Buffer);
    else
        printf(Buffer);
}
#endif

void CFormDisplay::Display()
{
    DisplayUnindexed();

    DisplayProlog();

    CObjPtr<DDocument> Doc1 = Mirror.NewFirstDoc(&Result);
    while (Doc1 != NULL)
    {   CObjPtr<DDocument> Doc2 = Mirror.NewNextDoc(&Result, Doc1);
	   DisplayDoc(Doc1.Drop());
	   Doc1 = Doc2.Drop();
    }

    DisplayEpilog();
}

void CFormDisplay::DisplayUnindexed()
{
    DISArray<DString> &Unknown = Result.GetUnknown();
    if (Unknown.GetCount())
    {   DString NotFoundTemplate = Options.Get(O_NFLINE_1);

	   ErrPrint(Options.Get(T_NOTFOUNDHEADER_N));
	   DISArray<DString>::DIterator Itr = Unknown.InitIterator();
	   while (Unknown.IsDereferenceable(Itr))
	   {	  DString *CurStr = Unknown.GetCurrent(Itr);
		  Assert(!CurStr->IsEmpty());
                  CFormat Format;
		  Format.Add(*CurStr);
		  ErrPrint(Format.GetMessage(NotFoundTemplate));
		  ++Itr;
	   }
   }
}

void CFormDisplay::DisplayProlog()
{
    CFormat Format;
    OutPrint(Format.GetMessage(Options.Get(O_PROLOG_0)));
}

#if defined(DEBUG_epilog)
int Index = 0;
#endif
void CFormDisplay::DisplayEpilog()
{
    CFormat Format;
    Format.Add(GetColorString(Options.COptions<int>::Get(O_DEFAULTCOLOR)));
    OutPrint(Format.GetMessage(Options.Get(O_EPILOG_1)));
}

void CFormDisplay::DisplayDoc(DDocument *InDoc)
{   CDocDesc DocDesc;
    const char *Template;

    CObjPtr<DDocument> Doc = InDoc;

    Mirror.GetDocName(Doc, DocDesc);

    CFormat Format;
    Format.Add(DocDesc.GetFileName());
    if (DocDesc.GetStartOffs() == 0)
        Template = O_DOCNAME_2;
    else
    {   Template = O_DOCNAME_3;
        Format.Add(DocDesc.GetStartOffs());
    }
    Format.Add(GetColorString(Options.COptions<int>::Get(O_DOCNAMECOLOR)));
    OutPrint(Format.GetMessage(Options.Get(Template)));

    CRelevance *Relevance = Doc->GetRelObj();
    if (Relevance)
        DisplayRelevance(Relevance, Doc->GetRelRes());

    // optimization
    if (!SkipClots)
	   DisplayDocContents(Doc.Drop());

    CFormat EndFormat;
    OutPrint(EndFormat.GetMessage(Options.Get(O_DOCEND)));
}

void CFormDisplay::DisplayDocContents(DDocument *InDoc)
{   
    try
    {   CClot *Clot;

	   CObjPtr<DDocument> Doc = InDoc;
	   CObjPtr<CClotCursor> ClotCursor = Mirror.NewClotCursor(Doc.Drop(),
            Options.COptions<int>::Get(O_SPOTDIST));
        int HighlightCount = 0;
        while ((Clot = ClotCursor->GetCurrent()) != NULL)
        {   if (DisplayClot(Clot, HighlightCount))
                break;
            Mirror.Iterate(ClotCursor);
        }
    }
    catch (CPassageOverflow &)
    {   CFormat Format;
        Format.Add(DString::MaxAllowedLength);
        OutPrint(Format.GetMessage(Options.Get(O_PASSAGEOVERFLOW_1)));
    }
}

void CFormDisplay::DisplayRelevance(CRelevance *Relevance, double RelRes)
{
    for (CRelevance::EFeature Feature = (CRelevance::EFeature)0;
        Feature < CRelevance::FeatureCount;
        Feature = (CRelevance::EFeature)(Feature + 1))
    {   CFormat Format;
        Format.Add(CRelevance::FeatureName[Feature]);
        Format.Add(Relevance->GetMetric(Feature));
        OutPrint(Format.GetMessage(Options.Get(O_METRIC_2)));
    }
    CFormat Format;
    Format.Add(RelRes);
    OutPrint(Format.GetMessage(Options.Get(O_RELEVANCE_1)));
}

int CFormDisplay::DisplayClot(CClot *Clot, int &HighlightCount)
{   int ColorCode[2];
    int HeadTailColor;
    int HighlightMax;
    DString Passage;

    ColorCode[0] = Options.COptions<int>::Get(O_TEXTCOLOR);
    ColorCode[1] = Options.COptions<int>::Get(O_OCCCOLOR);
    HeadTailColor = Options.COptions<int>::Get(O_CLOTCOLOR);
    HighlightMax = Options.COptions<int>::Get(O_HIGHLIGHTMAX);

    CFormat Format;
    Format.Add(GetColorString(HeadTailColor));
    OutPrint(Format.GetMessage(Options.Get(O_CLOTHEAD_1)));

    int ColorIndex = Clot->GetFirstLow() ? 0 : 1;
    int i = 0;
    while (!Clot->GetPassage(i, Passage))
    {   if (ColorIndex == 1 && HighlightMax > 0)
        {   HighlightCount++;
            if (HighlightCount > HighlightMax)
            {   // reuse Format
		        OutPrint(Format.GetMessage(Options.Get(O_CLOTBREAK_1)));
                return 1;
            }
        }

        DisplayPassage(Passage,
            ColorIndex ? O_HIGHPASS_4 : O_LOWPASS_4,
            ColorCode[ColorIndex]);

	   ColorIndex = ColorIndex ? 0 : 1;
	   i++;
    }

    // reuse Format
    OutPrint(Format.GetMessage(Options.Get(O_CLOTTAIL_1)));
    return 0;
}

// O_FORMATSTRING no longer supported
void CFormDisplay::DisplayPassage(const DString &Passage, 
    const char *Template,
    int ColorCode)
{	
    DString Before;
	DString After;
    DString ParsedPassage = Passage;
    PunctParse->ParsePunct(ParsedPassage, Before, After);
#if defined(DEBUG_passage)
    printf("#%s# -> #%s# + #%s# + #%s#\n", 
        Passage.c_str(), 
        ParsedPassage.c_str(),
        Before.c_str(),
        After.c_str());
#endif

    CFormat Format;
    Format.Add(Before);
    Format.Add(ParsedPassage);
    Format.Add(After);
    Format.Add(GetColorString(ColorCode));
    OutPrint(Format.GetMessage(Options.Get(Template)));
}
/*void CAnsiDisplay::DisplayPassage(const DString &Passage, int ColorCode)
{   char *ExFormat;

    DString FormatString = Options.Get(O_FORMATSTRING);
    if (FormatString.IsEmpty())
    {   SetColor(ColorCode);
	   printf("%s", Passage.c_str());
	   return;
    }
    int FormatCode = Options.COptions<int>::Get(O_FORMATCOLOR);

    CMemPtr<char> PassageBuf(Passage.GetLength() + 1);
    strcpy((char *)PassageBuf, Passage.c_str());
    char *PassageTail = (char *)PassageBuf;
    const char *Format = FormatString.c_str();
    int FormatLen = FormatString.GetLength();
    while (*PassageTail != 0 &&
	   (ExFormat = strstr(PassageTail, Format)) != NULL)
    {   *ExFormat = 0;
	   SetColor(ColorCode);
	   if (*PassageTail != 0)
		  printf("%s", PassageTail);
	   SetColor(FormatCode);
	   printf("%s", Format);
	   PassageTail = ExFormat + FormatLen;
    }
    SetColor(ColorCode);
    printf("%s", PassageTail);
}*/

int CFormDisplay::IsEmpty(const char *Name)
{
    DString Value = Options.Get(Name);
    return Value.IsEmpty();
}
