#include "myfndt.h"

#include "grazeopt.h"
#include "fillopt.h"
#include "omirror.h"
#include "ocolors.h"
#include "obase.h"
#include "ooutform.h"
#include "omode.h"
#include "otmp.h"
#include "ogrrpt.h"
#include "oquery.h"
#include "osoftop.h"
#include "oweb.h"

#include "mirror.h"
#include "docum.h"
#include "spot.h"
#include "txtcolor.h"
#include "clotcurs.h"
#include "clot.h"
#include "result.h"
#include "relevanc.h"
#include "wordlist.h"
#include "txtrpt.h"
#include "fbdual.h"
#include "decompr.h"
#include "docdesc.h"
#include "dispfac.h"
#include "exedrive.h"
#include "term.h"
#include "query.h"
#include "softop.h"
#include "qbuild.h"
#include "timerep.h"
#include "figure.h"
#include "stone.h"
#include "unquote.h"

#include <stdio.h>
#include <string.h>

int WDLimit = 5;
int WDDelta = 5;

void Run(int Count, char *Arg[]);
void RunWordList(CMirror &Mirror, CGrazerOptions &Options);
void RunFigure(CMirror &Mirror, CGrazerOptions &Options);
void RunSearch(CMirror &Mirror,
	CFormula &Formula,
	CGrazerOptions &Options);
void RunComplexQuery(CMirror &Mirror, 
    const DString &QueryFile, 
    CGrazerOptions &Options);

void Print(const char *Str);

void Run(int Count, char *Arg[])
{
	Assert(Arg != NULL);
	CGrazerOptions Options(GetExeDrive(Arg[0]));
	if (FillOptions(Options))
		return;
	CMirror Mirror(new CDecompressor(Options), Options);

    DString WordListFile = Options.Get(O_WORDLISTFILE);
    if (!WordListFile.IsEmpty())
    {   DString FrgFile = Options.Get(O_FRGFILE);
	   if (FrgFile.IsEmpty())
		  RunWordList(Mirror, Options);
	   else
		  RunFigure(Mirror, Options);
    }

    DString QueryFile = Options.Get(O_QEXPR);
    if (!QueryFile.IsEmpty())
	{	if (Count > 1)
		    printf("warning: command-line args overriding stored query\n");
		else
        {   RunComplexQuery(Mirror, QueryFile, Options);
		    return;
		}
	}

	if (Count > 1)
	{	if (Count == 2 && Options.COptions<int>::Get(O_CGIPARAM))
        {   CUnquote Unquote(Options.Get(O_TEMPDIR), Arg[1]);
            if (!Unquote.IsValid())
                printf("error: %s is not a valid string\n", Arg[1]);
            else
                RunComplexQuery(Mirror, Unquote.GetQueryFile(), Options);
        }
        else
        {   CTerm Term(Count, Arg, Print, Options);
            RunSearch(Mirror, Term, Options);
        }
		return;
	}

	if (WordListFile.IsEmpty())
		printf("specify words to search for\n");
}

void RunSearch(CMirror &Mirror,
	CFormula &Formula,
	CGrazerOptions &Options)
{
    CTimeReporter Reporter(4, Options.Get(T_SEARCHPROGRESS_DDN));

    printf("searching...\n");

    int BreakFlag = 0;
    CObjPtr<CResult> Result = Mirror.NewResult(Formula,
	   Options.COptions<int>::Get(O_DOCLIMIT),
	   Options.Get(O_TEMPDIR),
	   Reporter,
	   BreakFlag);
    Assert(Result != NULL);

    CDisplayFacade Display(Mirror, *Result, Options);
    Display.Display();
}

void RunWordList(CMirror &Mirror, CGrazerOptions &Options)
{	DString CurWord;
	long DocCount;
	long OccCount;

	CTextReporter Reporter("%d: creating word list (%d%%)\n",
		100);
	CObjPtr<CWordList> WordList = Mirror.NewWordList(Options,
		Reporter);

	DString WLName = Options.Get(O_WORDLISTFILE);
	Assert(!WLName.IsEmpty());
	CTextReporter ExportReporter("%d: exporting word list (%d%%)\n",
		100);
	CFileObj WLOut(WLName.c_str(), "w+b");
	CNewKey Key1(WordList->NewFirstKey());
	long TotalCount = WordList->GetWordCount();
	long CurCount = 1;
	while (Key1 != NULL)
	{    WordList->GetWord(Key1, CurWord, DocCount, OccCount);
		WLOut.Print("%s\t%ld\t%ld\n",
			CurWord.c_str(),
			DocCount,
			OccCount);
		ExportReporter.ReportProgress(1, (int)((100 * CurCount) / TotalCount));
		CNewKey Key2(WordList->NewNextKey(Key1));
		Key1 = Key2.Drop();
		CurCount++;
	}
}

void RunFigure(CMirror &Mirror, CGrazerOptions &Options)
{
    CTimeReporter Reporter(4, "%d: creating word figure (%d%%)\n");
    CObjPtr<CFigure> Figure = Mirror.NewFigure(Print, Options, Reporter);
    if (!Figure)
	   return;

    DString FName = Options.Get(O_WORDLISTFILE);
    Assert(!FName.IsEmpty());
    printf("writing figure into %s...\n", FName.c_str());
    CTimeReporter ExportReporter(4, "%d: exporting word figure (%d%%)\n");
    CFileObj FOut(FName.c_str(), "w+b");
    CLockedKey Key1(Figure->LockFirstKey(), Figure);
    long TotalCount = Figure->GetStoneCount();
    long CurCount = 1;
    while (Key1 != NULL)
    {   CStone Stone = Figure->GetStone(Key1);
	   Stone.PrintLine(FOut);
	   ExportReporter.ReportProgress(1, (int)((100 * CurCount) / TotalCount));
	   CLockedKey Key2(Figure->LockNextKey(Key1), Figure);
	   Key1 = Key2.Drop();
	   CurCount++;
    }
    printf("Done.\n");
}

void Print(const char *Str)
{
    printf("%s", Str);
}

void RunComplexQuery(CMirror &Mirror, 
    const DString &QueryFile, 
    CGrazerOptions &Options)
{   
    if (FillUno(Options, O_SOFTOPS, CSoftOperator::SetUnoClass))
	    return;

	CQueryBuilder QueryBuilder(Print, Options);
	CFormula *FreeFormula;
	char *err = QueryBuilder.Build(QueryFile, &FreeFormula);
	if (err)
	{   printf("%s\n", err);
        free(err);
        return;
    }
	CObjPtr<CFormula> Formula = FreeFormula;

	RunSearch(Mirror, *FreeFormula, Options);
}
