#if !defined(__mirror_h)
#define __mirror_h

#include "gindex.h"
#include "grazeopt.h"
#include "tl.h"
#include "cluster.h"
#include "filebase.h"
#include "docdesc.h"
#include "rxw.h"

class CNumClot;
class CClotCursor;
class CClot;
class CReader;
class CSpot;
class CSoftOperator;
class CResult;
class CDocWrapper;
class CWordList;
class CReporter;
class CDecompressor;
class CInterval;
class CFrontLine;
class CExtras;
class DDocument;
class CFormula;
class CFigure;
class CStone;
class CFileBufRead;
class CWordProvider;
class CInfrastructure;

class CMirror
{
private:
    CFileBase FileBase;
    CGrazeIndex GrazeIndex;
    CBook ClusterBook;
    const DString Separators;
    const CRegex Divider;
    const int ReaderMargin;
    CDecompressor *Decompressor;
    int HeadLength;
    CInterval *Interval;
    CInfrastructure *Infrastructure;

public:
	CMirror(CDecompressor *aDecompressor, // owned by this
		CGrazerOptions &Options);
	virtual ~CMirror();

	CResult *NewResult(CFormula &Formula,
		int DocLimit,
		const DString &ResultDir,
		CReporter &Reporter,
		int &BreakFlag);

	DDocument *NewFirstDoc(CResult *Result);
	DDocument *NewNextDoc(CResult *Result, DDocument *Doc);

	CSpot *NewFirstSpot(DDocument *Doc);
	CSpot *NewNextSpot(CSpot *Spot, DDocument *Doc);

	CClotCursor *NewClotCursor(DDocument *InDoc, int Distance);
		// takes ownership of InDoc
	void Iterate(CClotCursor *ClotCursor);
		// asserts ClotCursor->GetCurrent

// not multithread-safe
#if !defined(_MSC_VER) 
	CWordList *NewWordList(CGrazerOptions &Options,
		CReporter &Reporter);
     CFigure *NewFigure(void (*Print)(const char *),
		CGrazerOptions &Options,
		CReporter &Reporter);
#endif

	static short GetPerc(const DString &CurWord);

// functions which must be synchronized
	virtual void GetDocName(DDocument *Doc, CDocDesc &DocDesc);

	virtual void DropDecompressedFiles();

protected:
	virtual CSpot *NewSpot(ulong DocNum, 
		ulong LocalOcc, 
		long Iterator);

	virtual short IterateBack(ulong DocNum, CCluster &Cluster);
		// Cluster is in & out param
		// returns 0 OK (cluster was changed), 1 cluster is first
		// in document

	// takes ownership of InDoc
	virtual CDocWrapper *NewDocWrapper(DDocument *InDoc);

	virtual void GetDocCount(CFixedOcc *FixedOcc,
		long &DocTotal,
		long &OccCount);

	virtual CDatumEnvelopeTemplate<CFixedOcc> GetOccList(
		const DString &IndexedWord);

	virtual ulong GetTotalDocCount();

    virtual ulong AddCachedDoc(CFrontLine &FrontLine, ulong GlobOcc);

	virtual int IsUnder(ulong Num);
	virtual int IsOver(ulong Num);

private:
     CResult *NewSearchResult(CFrontLine &FrontLine,
		CExtras &Extras);

     DDocument *NewFirstDoc(CFrontLine &FrontLine);
     DDocument *NewNextDoc(CFrontLine &FrontLine, DDocument *Doc);

	// takes ownership of InSpot
	CClot *NewClot(DDocument *Doc,
		CReader *Reader,
		int Distance,
		CSpot *InSpot,
		int WithUserHead);

	void FillClot(CClot *Clot,
		CReader *Reader,
		CNumClot *NumClot,
		int Distance,
		int WithUserHead);

	ushort FillClotHead(CClot *Clot,
		CReader *Reader,
		CNumClot *NumClot,
	    int Distance,
		int WithUserHead);

    CWordProvider *NewWordProvider(CGrazerOptions &Options, CReporter &Reporter);

    int AddForeground(CFigure &Figure,
	   const DString &FrgFile, 
	   void (*PrintErr)(const char *));
        // return 0 OK, 1 syntax error
	   // could use a reporter
    static int GetForegroundDoc(CFileBufRead &FrgFile, 
	   DString &FName, // out
	   long &Offs); // out
        // returns 0 OK, 1 EOF (FName & Offs might still be valid)
    int AddForegroundDoc(CFigure &Figure, 
	   const DString &FName,
	   long Offs,
	   void (*PrintErr)(const char *),
	   long Line);
    void AddStone(CFigure &Figure, 
	   const DString &SWord,
        CFixedOcc &FixedOcc);
    static CInfrastructure *NewInfrastructureOrNull(CGrazerOptions &Options);
};

#endif

