#include "myfndt.h"
#include "stone.h"
#include <math.h>

#define ASSERTVALID { Assert(ForegroundDocCount >= 0 && \
	   TotalDocCount > 0 && \
        TotalDocCount >= ForegroundDocCount && \
        ForegroundWordCount >= ForegroundDocCount && \
        TotalWordCount >= ForegroundWordCount); \
    Assert(!SWord.IsEmpty()); }

CStone::CStone(const DString &aSWord,
    long aForegroundDocCount,
    long aTotalDocCount,
    long aForegroundWordCount,
    long aTotalWordCount):
    ForegroundDocCount(aForegroundDocCount),
    TotalDocCount(aTotalDocCount),
    ForegroundWordCount(aForegroundWordCount),
    TotalWordCount(aTotalWordCount),
    SWord(aSWord)
{
    ASSERTVALID;
}

CStone::CStone(const CStone &Other):
    ForegroundDocCount(Other.ForegroundDocCount),
    TotalDocCount(Other.TotalDocCount),
    ForegroundWordCount(Other.ForegroundWordCount),
    TotalWordCount(Other.TotalWordCount),
    SWord(Other.SWord)
{
    ASSERTVALID;
}

CStone &CStone::operator = (const CStone &Other)
{
    if (&Other != this)
    {   ForegroundDocCount = Other.ForegroundDocCount;
	   TotalDocCount = Other.TotalDocCount;
	   ForegroundWordCount = Other.ForegroundWordCount;
	   TotalWordCount = Other.TotalWordCount;
	   SWord = Other.SWord;
	   ASSERTVALID;
    }
    return *this;
}

void CStone::PrintLine(CFileObj &FOut)
{
    FOut.Print("%s\t%f (%ld/%ld)\t%f\n",
        SWord.c_str(),
	   ((double)ForegroundDocCount)/TotalDocCount,
	   ForegroundDocCount,
        TotalDocCount,
        ((double)ForegroundWordCount)/TotalWordCount);
}

short CStone::Compare(const CStone &Other) const
{
    if (ForegroundDocCount != Other.ForegroundDocCount)
  	   return ForegroundDocCount < Other.ForegroundDocCount ? -1 : 1;

    if (TotalDocCount != Other.TotalDocCount)
  	   return TotalDocCount < Other.TotalDocCount ? -1 : 1;

/*    long ld = CheckedProduct(ForegroundDocCount, Other.TotalDocCount);
    long rd = CheckedProduct(Other.ForegroundDocCount, TotalDocCount);
    if (ld != rd)
	   return ld < rd ? -1 : 1;*/

    if (ForegroundWordCount != Other.ForegroundWordCount)
  	   return ForegroundWordCount < Other.ForegroundWordCount ? -1 : 1;

    if (TotalWordCount != Other.TotalWordCount)
  	   return TotalWordCount < Other.TotalWordCount ? -1 : 1;

/*    long lw = CheckedProduct(ForegroundWordCount, Other.TotalWordCount);
    long rw = CheckedProduct(Other.ForegroundWordCount, TotalWordCount);
    if (lw != rw)
	   return lw < rw ? -1 : 1;*/

    if (SWord < Other.SWord)
	   return -1;
    else
  	   return SWord == Other.SWord ? 0 : 1;
}

long CStone::CheckedProduct(long p1, long p2)
{
    Assert(p1 >= 0 && p2 >= 0);
    if (!p1 || !p2)
	   return 0;

    // if product overflows 32 bits (with reserve for rounding)
    if ((log(p1) + log(p2)) >= log(2) * 31)
	   throw CGeneralException("overflow: could not compare stones in figure");
    long p = p1 * p2;
    Assert(p >= p1 && p >= p2);
    return p;
}

