#include "myfndt.h"
#include "dynstem.h"

#define PROTO(rtype, name, par)		typedef rtype (*F##name)par;
#include "libstem.h"

#include "locusopt.h"
#include "ostem.h"
#include <dlfcn.h>

CDynamicStemmer::CDynamicStemmer(CLocusOptions &Options):
    Lib(NULL),
    Stemmer(NULL),
    ApplyProc(NULL),
    DoneProc(NULL)
{    
    try
    {   InitLib();
        InitInit(Options);
        InitApply();
        InitDone();
    }
    catch (...)
    {   Done();
        throw;
    }
}

CDynamicStemmer::~CDynamicStemmer()
{
    Done();
}

void CDynamicStemmer::Done()
{
    if (Lib)
    {   if (DoneProc && Stemmer)
            (reinterpret_cast<FStemDone>(DoneProc))(Stemmer);
        dlclose(Lib);
    }
}

int CDynamicStemmer::Apply(DString &WordForm)
{
    return (reinterpret_cast<FStemApply>(ApplyProc))(Stemmer, WordForm);
}

void CDynamicStemmer::InitLib()
{
    Lib = dlopen("libstem.so", RTLD_LAZY | RTLD_GLOBAL);
    if (!Lib)
    {   DString err = dlerror();
        printf("dlopen(\"libstem.so\") failed with %s\n", err.c_str());
        throw CGeneralException(err.c_str());
    }
}

void CDynamicStemmer::InitInit(CLocusOptions &Options)
{
    void *InitProc = dlsym(Lib, "StemInit");
    if (!InitProc)
    {   DString err = dlerror();
        printf("dlsym(\"StemInit\") failed with %s\n", err.c_str());
        throw CGeneralException(err.c_str());
    }

	Stemmer = (reinterpret_cast<FStemInit>(InitProc))(Options.Get(O_SDKEY),
	    Options.Get(O_SDSUPP),
	    Options.Get(O_SDDATA),
	    Options.Get(O_SDTABLE));
}

void CDynamicStemmer::InitApply()
{
    ApplyProc = dlsym(Lib, "StemApply");
    if (!ApplyProc)
    {   DString err = dlerror();
        printf("dlsym(\"StemApply\") failed with %s\n", err.c_str());
        throw CGeneralException(err.c_str());
    }
}

void CDynamicStemmer::InitDone()
{
    DoneProc = dlsym(Lib, "StemDone");
    if (!DoneProc)
    {   DString err = dlerror();
        printf("dlsym(\"StemDone\") failed with %s\n", err.c_str());
        throw CGeneralException(err.c_str());
    }
}
