#include "myfndt.h"
#include "watch.h"
#include "locusopt.h"
#include "change.h"

#include "osubdoc.h"
#include "oreader.h"

#include <stdio.h>

CWatch::CWatch(const CRegex &aDocBegin, short aMargin):
    Margin(0),
	DocBegin(aDocBegin), 
    PushMatch(4, 4),
    PopMatch(4, 4)
{
    if (!DocBegin.IsEmpty())
 		Margin = aMargin;
}	

CWatch::CWatch(CLocusOptions &Options):
    Margin(0),
	DocBegin(Options.Get(O_DIVIDER)), 
    PushMatch(4, 4),
    PopMatch(4, 4)
{
    if (!DocBegin.IsEmpty())
 		Margin = Options.COptions<int>::Get(O_READERMARGIN);
    
    CStringListValue List = Options.COptions<CStringListValue>::Get(O_DOCPARTS);
    CStringListValue::DIterator Itr = List.InitIterator();
    while (List.IsDereferenceable(Itr))
    {   DString Name = *Itr;
        PushMatch.Add(Options.COptions<DString>::GetUno(Name, O_PUSHPART));
        PopMatch.Add(Options.COptions<DString>::GetUno(Name, O_POPPART));
        Itr++;
    }
}

CWatch::~CWatch()
{
}

void CWatch::SetSubdocClass(CUnivOptions &Options,
	const CStringListValue &DocParts)
{   static const char *StringAttr[] = {O_PUSHPART, O_POPPART};

    Options.COptions<DString>::SetUnoClass(DocParts, StringAttr,
            sizeof(StringAttr) / sizeof(StringAttr[0]));
}

/* profile 3. 2. 1997: VIRUS-L, optimized build
with only document divider: 5 hr 43 min
with 3 doc parts (from, subject, quote): 6 hr 10 min */
void CWatch::Iterate(const char *Chunk, CChange &Change)
{   static const int PushIndex = 0;
    static const int PopIndex = 1;
    static const int PushMatch = 1;
    static const int PopMatch = 2;
    int From[2], To[2];
    short SubdocId;

    int HeadSize = Change.GetPosAfterSign();
    Change.ClearSign();

    const char *Tail = Chunk + HeadSize;
    int t = 0;
    if (GetPushMatch(Tail, From[PushIndex], To[PushIndex], SubdocId))
        t += PushMatch;
    short OldSubdoc = Change.GetChangedSubdoc();
    if (OldSubdoc && GetPopMatch(Tail, OldSubdoc, From[PopIndex], To[PopIndex]))
        t += PopMatch;

    if (!t)
        return;

    if (t == PushMatch + PopMatch)
    {   if (From[PushIndex] < From[PopIndex])
            t = PushMatch;
        else
        {   if (From[PushIndex] > To[PopIndex])
                t = PopMatch;
            else
            {   Change.Pop();
                Change.SetSign(From[PushIndex] + HeadSize, 
                    To[PushIndex] + HeadSize);
                Change.Push(SubdocId);
                return;
            }
        }
    }

    if (t == PushMatch)
    {   Change.SetSign(From[PushIndex] + HeadSize, To[PushIndex] + HeadSize);
        Change.Push(SubdocId);
    }
    else
    {   Change.SetSign(From[PopIndex] + HeadSize, To[PopIndex] + HeadSize);
        Change.Pop();
    }
}

short CWatch::GetValidSize(const char *Chunk, short ReadSize, int FromBegin)
{   int From, To, From2, To2;
    
    if (!DocBegin.Matches(Chunk, From, To))
        return ReadSize;

    if (!FromBegin || From > 0)
    {   Assert(From < ReadSize);
        return From;
    }

    if (To >= ReadSize)
        return ReadSize;

    if (!DocBegin.Matches(Chunk + To, From2, To2))
        return ReadSize;
    else
    {   Assert(To + From2 < ReadSize);
        return To + From2;
    }
}

int CWatch::GetPushMatch(const char *Chunk,
    int &From,
	int &To,
	short &SubdocId)
{   int CurrentFrom, CurrentTo;

    int ChunkLen = strlen(Chunk);
    int Count = PushMatch.GetCount();
    From = ChunkLen;
    for (int i = 0; i < Count; ++i)
    {   if (PushMatch[i].Matches(Chunk, CurrentFrom, CurrentTo))
        {   Assert(CurrentFrom < ChunkLen);
            if (CurrentFrom < From)
            {   From = CurrentFrom;
                To = CurrentTo;
                SubdocId = i + 1;
            }
        }
    }
    return From < ChunkLen;
}

int CWatch::GetPopMatch(const char *Chunk,
	short Subdoc,
    int &From,
	int &To)
{
    Assert(Subdoc > 0);
    return PopMatch[Subdoc - 1].Matches(Chunk, From, To);
}

short CWatch::GetMargin() const
{
    return Margin;
}
